/*
 * $Id: NdviProcessor.java,v 1.3 2005/03/25 00:13:58 hartmut Exp $
 *
 * Copyright (C) 2004 by Brockmann Consult (info@brockmann-consult.de)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation. This program is distributed in the hope it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.bc.beam.processor.ndvi;

import com.bc.progress.ProgressController;
import com.bc.progress.ProgressControllerPool;
import org.esa.beam.dataio.dimap.DimapProductConstants;
import org.esa.beam.dataio.envisat.EnvisatConstants;
import org.esa.beam.framework.dataio.ProductIO;
import org.esa.beam.framework.dataio.ProductWriter;
import org.esa.beam.framework.datamodel.Band;
import org.esa.beam.framework.datamodel.BitmaskDef;
import org.esa.beam.framework.datamodel.FlagCoding;
import org.esa.beam.framework.datamodel.MetadataAttribute;
import org.esa.beam.framework.datamodel.Product;
import org.esa.beam.framework.datamodel.ProductData;
import org.esa.beam.framework.processor.Processor;
import org.esa.beam.framework.processor.ProcessorConstants;
import org.esa.beam.framework.processor.ProcessorException;
import org.esa.beam.framework.processor.ProductRef;
import org.esa.beam.framework.processor.Request;
import org.esa.beam.framework.processor.ui.ProcessorUI;
import org.esa.beam.util.ProductUtils;

import java.awt.Color;
import java.io.File;
import java.io.IOException;
import java.util.logging.Logger;

/**
 * The <code>NdviProcessor</code> implements all specific functionality to calculate a processor
 * product from a given MERIS Level 1b product. This simple processor does not take any flags into
 * account, it just calculates the processor over the whole product.
 */
public class NdviProcessor extends Processor {

    // Constants
    public static final String PROCESSOR_NAME = "NDVI Processor";
    public static final String PROCESSOR_VERSION = "1.0";
    public static final String PROCESSOR_COPYRIGHT = "Copyright (C) 2003 by Brockmann Consult (info@brockmann-consult.de)";

    public static final String LOGGER_NAME = "beam.processor.processor";

    public final static String DEFAULT_OUTPUT_DIR_NAME = "processor";
    public final static String DEFAULT_OUTPUT_FORMAT = DimapProductConstants.DIMAP_FORMAT_NAME;
    public static final String DEFAULT_OUTPUT_PRODUCT_NAME = "MER_NDVI2P";

    public static final String REQUEST_TYPE = "NDVI";

    public static final String NDVI_PRODUCT_TYPE = "MER_NDVI2P";
    public static final String NDVI_BAND_NAME = "ndvi";
    public static final String NDVI_FLAGS_BAND_NAME = "ndvi_flags";
    public static final String NDVI_ARITHMETIC_FLAG_NAME = "NDVI_ARITHMETIC";
    public static final String NDVI_LOW_FLAG_NAME = "NDVI_NEGATIVE";
    public static final String NDVI_HIGH_FLAG_NAME = "NDVI_SATURATION";
    public static final int NDVI_ARITHMETIC_FLAG_VALUE = 1 << 0;
    public static final int NDVI_LOW_FLAG_VALUE = 1 << 1;
    public static final int NDVI_HIGH_FLAG_VALUE = 1 << 2;

    public static final String LOWER_INPUT_BAND_NAME = "radiance_6";
    public static final String UPPER_INPUT_BAND_NAME = "radiance_10";
    public static final String L1FLAGS_INPUT_BAND_NAME = "l1_flags";

//    public static final String LON_GRID_NAME = EnvisatConstants.LAT_DS_NAME;
//    public static final String LAT_GRID_NAME = EnvisatConstants.LON_DS_NAME;

    // Fields
    private Product _inputProduct;
    private Product _outputProduct;
    private Band _lowerInputBand;
    private Band _upperInputBand;
    private Band _l1FlagsInputBand;
    private Band _ndviOutputBand;
    private Band _l1FlagsOutputBand;
    private Band _ndviFlagsOutputBand;
    private Logger _logger;

    public NdviProcessor() {
        _logger = Logger.getLogger(LOGGER_NAME);
    }

    /**
     * Worker method invoked by framework to process a single request.
     */
    public void process() throws ProcessorException {

        try {
            _logger.info("Started processing ...");

            // check the request type
            Request.checkRequestType(getRequest(), REQUEST_TYPE);

            // create progress bar and add it too the processor's listener list
//            useProgressBar();

            // load input product
            loadInputProduct();

            // create the output product
            createOutputProduct();

            // and process the processor
            processNdvi();
        } catch (IOException e) {
            // catch all exceptions expect ProcessorException and throw ProcessorException
            throw new ProcessorException(e.getMessage());
        }
    }

    /**
     * Retrieves the name of the processor
     */
    public String getName() {
        return PROCESSOR_NAME;
    }

    /**
     * Retrieves a version string of the processor
     */
    public String getVersion() {
        return PROCESSOR_VERSION;
    }

    /**
     * Retrieves copyright information of the processor
     */
    public String getCopyrightInformation() {
        return PROCESSOR_COPYRIGHT;
    }

    /**
     * Creates the UI for the processor. Override to perform processor specific
     * UI initializations.
     */
    public ProcessorUI createUI() throws ProcessorException {
        return new NdviProcessorUI();
    }

    ///////////////////////////////////////////////////////////////////////////
    /////// END OF PUBLIC
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Loads the input product from the request. Opens the product and opens
     * both bands needed to process the processor.
     */
    private void loadInputProduct() throws ProcessorException, IOException {
        _inputProduct = loadInputProduct(0);

        if (!EnvisatConstants.MERIS_RR_L1B_PRODUCT_TYPE_NAME.equals(_inputProduct.getProductType())
                && !EnvisatConstants.MERIS_FR_L1B_PRODUCT_TYPE_NAME.equals(_inputProduct.getProductType())) {
            throw new ProcessorException("Invalid product type: MERIS Level 1b required.");
        }

        // try to retrieve the bands needed
        // --------------------------------
        _lowerInputBand = _inputProduct.getBand(LOWER_INPUT_BAND_NAME);
        if (_lowerInputBand == null) {
            throw new ProcessorException("Can not load band " + LOWER_INPUT_BAND_NAME);
        }
        _logger.info(ProcessorConstants.LOG_MSG_LOADED_BAND + LOWER_INPUT_BAND_NAME);

        _upperInputBand = _inputProduct.getBand(UPPER_INPUT_BAND_NAME);
        if (_upperInputBand == null) {
            throw new ProcessorException("Can not load band " + UPPER_INPUT_BAND_NAME);
        }
        _logger.info(ProcessorConstants.LOG_MSG_LOADED_BAND + UPPER_INPUT_BAND_NAME);

        _l1FlagsInputBand = _inputProduct.getBand(L1FLAGS_INPUT_BAND_NAME);
        if (_upperInputBand == null) {
            throw new ProcessorException("Can not load band " + L1FLAGS_INPUT_BAND_NAME);
        }
        _logger.info(ProcessorConstants.LOG_MSG_LOADED_BAND + L1FLAGS_INPUT_BAND_NAME);
    }

    /**
     * Creates the output product skeleton.
     */
    private void createOutputProduct() throws ProcessorException,
                                              IOException {
        // get the request from the base class
        // -----------------------------------
        Request request = getRequest();

        // get the scene size from the input product
        // -----------------------------------------
        int sceneWidth = getSceneWidth();
        int sceneHeight = getSceneHeight();

        // get the output product from the request. The request holds objects of
        // type ProductRef which contain all the information needed here
        // --------------------------------------------------------------------
        ProductRef outputRef;

        // the request can contain any number of output products, we take the first ..
        outputRef = request.getOutputProductAt(0);
        if (outputRef == null) {
            throw new ProcessorException("No output product in request");
        }

        // create the in memory represenation of the output product
        // ---------------------------------------------------------
        // the product itself
        _outputProduct = new Product(DEFAULT_OUTPUT_PRODUCT_NAME, NDVI_PRODUCT_TYPE, sceneWidth, sceneHeight);

        // create and add the NDVI band
        //
        _ndviOutputBand = new Band(NDVI_BAND_NAME, ProductData.TYPE_UINT8, sceneWidth, sceneHeight);
        _ndviOutputBand.setScalingOffset(0.0);
        _ndviOutputBand.setScalingFactor(1.0 / 255.0);
        _outputProduct.addBand(_ndviOutputBand);

        // copy all tie point grids to output product
        //
        ProductUtils.copyTiePointGrids(_inputProduct, _outputProduct);

        // copy geo-coding and the lat/lon tiepoints to the output product
        //
        ProductUtils.copyGeoCoding(_inputProduct, _outputProduct);

        // copy L1b flag band
        //
        ProductUtils.copyFlagBands(_inputProduct, _outputProduct);
        _l1FlagsOutputBand = _outputProduct.getBand(L1FLAGS_INPUT_BAND_NAME);

        // create and add the NDVI flags coding
        //
        FlagCoding ndviFlagCoding = createNdviFlagCoding();
        _outputProduct.addFlagCoding(ndviFlagCoding);

        // create and add the NDVI flags band
        //
        _ndviFlagsOutputBand = new Band(NDVI_FLAGS_BAND_NAME, ProductData.TYPE_UINT8, sceneWidth, sceneHeight);
        _ndviFlagsOutputBand.setDescription("NDVI specific flags");
        _ndviFlagsOutputBand.setFlagCoding(ndviFlagCoding);
        _outputProduct.addBand(_ndviFlagsOutputBand);

        // Copy predefined bitmask definitions
        ProductUtils.copyBitmaskDefs(_inputProduct, _outputProduct);
        _outputProduct.addBitmaskDef(new BitmaskDef(NDVI_ARITHMETIC_FLAG_NAME.toLowerCase(),
                                                    "An arithmetic exception occured.",
                                                    NDVI_FLAGS_BAND_NAME + "." + NDVI_ARITHMETIC_FLAG_NAME,
                                                    Color.red.brighter(), 0.7f));
        _outputProduct.addBitmaskDef(new BitmaskDef(NDVI_LOW_FLAG_NAME.toLowerCase(),
                                                    "NDVI value is too low.",
                                                    NDVI_FLAGS_BAND_NAME + "." + NDVI_LOW_FLAG_NAME,
                                                    Color.red, 0.7f));
        _outputProduct.addBitmaskDef(new BitmaskDef(NDVI_HIGH_FLAG_NAME.toLowerCase(),
                                                    "NDVI value is too high.",
                                                    NDVI_FLAGS_BAND_NAME + "." + NDVI_HIGH_FLAG_NAME,
                                                    Color.red.darker(), 0.7f));


        // retrieve the default disk writer from the ProductIO package
        // this is the BEAM_DIMAP format, the toolbox native file format
        // and attach to the writer to the output product
        //
        ProductWriter writer = ProductIO.getProductWriter(ProductIO.DEFAULT_FORMAT_NAME);
        _outputProduct.setProductWriter(writer);

        // and initialize the disk represenation
        //
        writer.writeProductNodes(_outputProduct, new File(outputRef.getFilePath()));

        _logger.info("Output product successfully created");
    }

    private int getSceneHeight() {
        return _inputProduct.getSceneRasterHeight();
    }

    private int getSceneWidth() {
        return _inputProduct.getSceneRasterWidth();
    }

    public static FlagCoding createNdviFlagCoding() {

        FlagCoding ndviFlagCoding = new FlagCoding("ndvi_flags");
        ndviFlagCoding.setDescription("NDVI Flag Coding");

        MetadataAttribute attribute;

        attribute = new MetadataAttribute(NDVI_ARITHMETIC_FLAG_NAME, ProductData.TYPE_INT32);
        attribute.getData().setElemInt(NDVI_ARITHMETIC_FLAG_VALUE);
        attribute.setDescription("NDVI value calculation failed due to an arithmetic exception");
        ndviFlagCoding.addAttribute(attribute);

        attribute = new MetadataAttribute(NDVI_LOW_FLAG_NAME, ProductData.TYPE_INT32);
        attribute.getData().setElemInt(NDVI_LOW_FLAG_VALUE);
        attribute.setDescription("NDVI value is too low");
        ndviFlagCoding.addAttribute(attribute);

        attribute = new MetadataAttribute(NDVI_HIGH_FLAG_NAME, ProductData.TYPE_INT32);
        attribute.getData().setElemInt(NDVI_HIGH_FLAG_VALUE);
        attribute.setDescription("NDVI value is too high");
        ndviFlagCoding.addAttribute(attribute);

        return ndviFlagCoding;
    }


    /**
     * Performs the actual processing of the output product. Reads both input bands line
     * by line, calculates the processor and writes the result to the output band
     */
    private void processNdvi() throws IOException {

        int width = getSceneWidth();
        int height = getSceneHeight();

        // first of all - allocate memory for single scan lines
        //
        float[] lower = new float[width];
        float[] upper = new float[width];
        float[] ndvi = new float[width];
        int[] ndviFlags = new int[width];
        int[] l1Flags = new int[width];

        // @todo make this values processing parameters
        //
        float upperFactor = 1.0f;
        float lowerFactor = 1.0f;

        float highValue, lowValue, ndviValue;
        int ndviFlagsValue;

        // Notify process listeners that processing has started
        //
        final ProgressController pc = ProgressControllerPool.getInstance().getProgressController();
        pc.fireProcessStarted("Generating NDVI pixels...", 0, height -1);

        try {
            // for all required bands loop over all scanlines
            //
            for (int y = 0; y < height; y++) {

                // read the input data
                //
                _lowerInputBand.readPixels(0, y, width, 1, lower);
                _upperInputBand.readPixels(0, y, width, 1, upper);
                _l1FlagsInputBand.readPixels(0, y, width, 1, l1Flags);

                // process the complete scanline
                //
                for (int x = 0; x < width; x++) {
                    highValue = upperFactor * upper[x];
                    lowValue = lowerFactor * lower[x];
                    ndviValue = (highValue - lowValue) / (highValue + lowValue);
                    ndviFlagsValue = 0;
                    if (Float.isNaN(ndviValue) || Float.isInfinite(ndviValue)) {
                        ndviFlagsValue |= NDVI_ARITHMETIC_FLAG_VALUE;
                        ndviValue = 0f;
                    }
                    if (ndviValue < 0.0f) {
                        ndviFlagsValue |= NDVI_LOW_FLAG_VALUE;
                        //ndviValue = 0.0f;
                    }
                    if (ndviValue > 1.0f) {
                        ndviFlagsValue |= NDVI_HIGH_FLAG_VALUE;
                        //ndviValue = 1.0f;
                    }
                    ndvi[x] = ndviValue;
                    ndviFlags[x] = ndviFlagsValue;
                }

                // write the result
                //
                _ndviOutputBand.writePixels(0, y, width, 1, ndvi);
                _ndviFlagsOutputBand.writePixels(0, y, width, 1, ndviFlags);
                _l1FlagsOutputBand.writePixels(0, y, width, 1, l1Flags);

                // Notify process listeners about processing progress and
                // check whether or not processing shall be terminated
                //
                pc.fireProcessInProgress(y);
                if (pc.isTerminationRequested()) {
                    // Processing terminated!
                    // --> Completely remove output product
                    _outputProduct.getProductWriter().deleteOutput();
                    // Immediately terminate now
                    _logger.info(ProcessorConstants.LOG_MSG_PROC_CANCELED);
                    _logger.info("The output product is completely removed.");
                    return;
                }
            }
        } finally {
            pc.fireProcessEnded();
        }

        // If processing is not interupted the logger logs success
        //
        _logger.info(ProcessorConstants.LOG_MSG_SUCCESS);
    }
}
