/*
 * $Id: NdviProcessorUI.java,v 1.2 2005/03/23 18:20:20 hartmut Exp $
 *
 * Copyright (C) 2004 by Brockmann Consult (info@brockmann-consult.de)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation. This program is distributed in the hope it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.bc.beam.processor.ndvi;

import org.esa.beam.framework.dataio.ProductIO;
import org.esa.beam.framework.datamodel.Product;
import org.esa.beam.framework.param.ParamChangeEvent;
import org.esa.beam.framework.param.ParamChangeListener;
import org.esa.beam.framework.param.ParamGroup;
import org.esa.beam.framework.param.Parameter;
import org.esa.beam.framework.processor.ProcessorConstants;
import org.esa.beam.framework.processor.ProcessorException;
import org.esa.beam.framework.processor.ProductRef;
import org.esa.beam.framework.processor.Request;
import org.esa.beam.framework.processor.ui.AbstractProcessorUI;
import org.esa.beam.framework.processor.ui.ProcessorApp;
import org.esa.beam.framework.ui.GridBagUtils;
import org.esa.beam.util.Debug;
import org.esa.beam.util.Guardian;
import org.esa.beam.util.ProcessorUtils;

import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import java.awt.GridBagConstraints;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Vector;

public class NdviProcessorUI extends AbstractProcessorUI {

    private JTabbedPane _tabbedPane;
    private ParamGroup _paramGroup;
    private File _requestFile;
    private NdviRequestElementFactory _factory;

    /**
     * Creates the ui class with default parameters
     */
    public NdviProcessorUI() {
        _factory = NdviRequestElementFactory.getInstance();
    }

    /**
     * Retrieves the base component for the processor specific user interface classes. This can be any Java Swing
     * containertype. This method creates the UI from scratch if not present.
     */
    public JComponent getGuiComponent() {
        if (_tabbedPane == null) {
            createUI();
        }
        return _tabbedPane;
    }

    /**
     * Retrieves the requests currently edited.
     */
    public Vector getRequests() throws ProcessorException {
        Vector requests = new Vector();
        requests.add(createRequest());
        return requests;
    }

    /**
     * Sets a new Request list to be edited.
     *
     * @param requests the request list to be edited must not be <code>null</code>.
     */
    public void setRequests(Vector requests) throws ProcessorException {
        Guardian.assertNotNull("requests", requests);
        if (requests.size() > 0) {
            Request request = (Request) requests.elementAt(0);
            _requestFile = request.getFile();
            updateParamInputFile(request);
            updateParamOutputFile(request);
            updateParamOutputFormat(request);
        } else {
            setDefaultRequests();
        }
    }

    /**
     * Create a new default request for the sst processor and sets it to the UI
     */
    public void setDefaultRequests() throws ProcessorException {
        Vector requests = new Vector();
        requests.add(createDefaultRequest());
        setRequests(requests);
    }

    /**
     * Sets the processor app for the UI
     */
    public void setApp(ProcessorApp app) {
    }

    ///////////////////////////////////////////////////////////////////////////
    /////// END OF PUBLIC
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Creates a request with all paramneters set to their respective default values.
     *
     * @return the default request
     */
    private Request createDefaultRequest() {
        _paramGroup.getParameter(ProcessorConstants.INPUT_PRODUCT_PARAM_NAME).setDefaultValue();
        _paramGroup.getParameter(ProcessorConstants.OUTPUT_PRODUCT_PARAM_NAME).setDefaultValue();
        _paramGroup.getParameter(ProcessorConstants.OUTPUT_FORMAT_PARAM_NAME).setDefaultValue();
        return createRequest();
    }

    /**
     * Creates all user interface components of the sst user interface
     */
    private void createUI() {
        initParamGroup();
        _tabbedPane = new JTabbedPane();
        _tabbedPane.add("I/O Parameters", createPathTab());
    }

    /**
     * Initializes theparameter group to hold all parameter needed for the processor.
     */
    private void initParamGroup() {
        _paramGroup = new ParamGroup();
        final Parameter inputProductParameter = _factory.createDefaultInputProductParameter();
        _paramGroup.addParameter(inputProductParameter);
        _paramGroup.addParameter(_factory.createDefaultOutputProductParameter());
        _paramGroup.addParameter(_factory.createOutputFormatParameter());

        inputProductParameter.addParamChangeListener(new ParamChangeListener() {
            public void parameterValueChanged(ParamChangeEvent event) {
                checkForValidInputProduct(inputProductParameter);
            }
        });
    }

    /**
     * Create the ui tab for the path editing.
     *
     * @return the panel containing the paths tab
     */
    private JPanel createPathTab() {
        JPanel panel = GridBagUtils.createDefaultEmptyBorderPanel();
        final GridBagConstraints gbc = GridBagUtils.createConstraints(null);
        gbc.gridy = 0;
        Parameter param;

        // input product
        // -------------
        param = _paramGroup.getParameter(ProcessorConstants.INPUT_PRODUCT_PARAM_NAME);
        gbc.gridy++;
        GridBagUtils.setAttributes(gbc, "anchor=SOUTHWEST, weighty=1");
        GridBagUtils.addToPanel(panel, param.getEditor().getLabelComponent(), gbc);
        gbc.gridy++;
        GridBagUtils.setAttributes(gbc, "anchor=NORTHWEST, fill=HORIZONTAL, weightx=1, weighty=1");
        GridBagUtils.addToPanel(panel, param.getEditor().getComponent(), gbc);

        // output product
        // --------------
        param = _paramGroup.getParameter(ProcessorConstants.OUTPUT_PRODUCT_PARAM_NAME);
        gbc.gridy++;
        GridBagUtils.setAttributes(gbc, "insets.top=7,anchor=SOUTHWEST, weighty=0.5");
        GridBagUtils.addToPanel(panel, param.getEditor().getLabelComponent(), gbc);
        gbc.gridy++;
        GridBagUtils.setAttributes(gbc, "insets.top=0,anchor=NORTHWEST, weighty=1");
        GridBagUtils.addToPanel(panel, param.getEditor().getComponent(), gbc);

        // output format
        // -------------
        param = _paramGroup.getParameter(ProcessorConstants.OUTPUT_FORMAT_PARAM_NAME);
        gbc.gridy++;
        GridBagUtils.setAttributes(gbc, "insets.top=7, anchor=SOUTHWEST, fill=NONE, weightx = 0, weighty=0.5");
        GridBagUtils.addToPanel(panel, param.getEditor().getLabelComponent(), gbc);
        gbc.gridy++;
        GridBagUtils.setAttributes(gbc, "insets.top=0, anchor=NORTHWEST, weighty=0.5");
        GridBagUtils.addToPanel(panel, param.getEditor().getComponent(), gbc);

        return panel;
    }

    private Request createRequest() {
        Request request = new Request();
        request.setType(NdviProcessor.REQUEST_TYPE);
        request.setFile(_requestFile);
        request.addInputProduct(createInputProductRef());
        request.addOutputProduct(createOutputProductRef());
        request.addParameter(createOutputFormatParamForRequest());
        return request;
    }

    /**
     * Creates an input product reference from the request.
     *
     * @return
     */
    private ProductRef createInputProductRef() {
        try {
            final String filePath = _paramGroup.getParameter(ProcessorConstants.INPUT_PRODUCT_PARAM_NAME).getValueAsText();
            final URL url = new URL("file", "", filePath);
            if (url != null) {
                return new ProductRef(url, null, null);
            }
        } catch (MalformedURLException e) {
            Debug.trace(e);
        }
        return null;
    }

    /**
     * Creates an output product reference from the request.
     *
     * @return
     */
    private ProductRef createOutputProductRef() {
        final String fileName = _paramGroup.getParameter(ProcessorConstants.OUTPUT_PRODUCT_PARAM_NAME).getValueAsText();
        final String fileFormat = _paramGroup.getParameter(ProcessorConstants.OUTPUT_FORMAT_PARAM_NAME).getValueAsText();

        return ProcessorUtils.createProductRef(fileName, fileFormat);
    }

    private Parameter createOutputFormatParamForRequest() {
        String outputFormat = _paramGroup.getParameter(ProcessorConstants.OUTPUT_FORMAT_PARAM_NAME).getValueAsText();
        return new Parameter(ProcessorConstants.OUTPUT_FORMAT_PARAM_NAME, outputFormat);
    }

    private void updateParamOutputFormat(Request request) {
        String format = request.getParameter(ProcessorConstants.OUTPUT_FORMAT_PARAM_NAME).getValueAsText();
        _paramGroup.getParameter(ProcessorConstants.OUTPUT_FORMAT_PARAM_NAME).setValue(format, null);
    }

    private void updateParamOutputFile(Request request) {
        File file = new File(request.getOutputProductAt(0).getFilePath());
        _paramGroup.getParameter(ProcessorConstants.OUTPUT_PRODUCT_PARAM_NAME).setValue(file, null);
    }

    private void updateParamInputFile(Request request) {
        File file = new File(request.getInputProductAt(0).getFilePath());
        _paramGroup.getParameter(ProcessorConstants.INPUT_PRODUCT_PARAM_NAME).setValue(file, null);
    }

    /**
     * Brings up a message box if the input product is not valid. Valid input products are: products wich contains at
     * least the bands named '{@link NdviProcessor#LOWER_INPUT_BAND_NAME lowBandName}' and '{@link
     * NdviProcessor#UPPER_INPUT_BAND_NAME highBandName}'. The message box only comes up if the parameter contains an
     * existing file. So you can create requests with not existing input products without interfering message Box.
     *
     * @param parameter
     */
    private void checkForValidInputProduct(Parameter parameter) {
        Object value = parameter.getValue();
        File file = null;
        if (value instanceof File) {
            file = (File) value;
        }
        if (value instanceof String) {
            file = new File((String) value);
        }
        if (file == null || !file.exists()) {
            return;
        }
        String msg = null;
        try {
            Product product = ProductIO.readProduct(file, null);
            if (product != null) {
                final String lowBandName = NdviProcessor.LOWER_INPUT_BAND_NAME;
                final String highBandName = NdviProcessor.UPPER_INPUT_BAND_NAME;
                if (product.getBand(highBandName) == null || product.getBand(lowBandName) == null) {
                    msg = "The NDVI Processor only works with products which contains\n" +
                          "at least the bands '" + lowBandName + "' and '" + highBandName + "'.";
                }
            } else {
                msg = "Unknown file format.";
            }
        } catch (IOException e) {
            msg = e.getMessage();
        }
        if (msg != null) {
            JOptionPane.showMessageDialog(getGuiComponent(), "Invalid input file:\n" + msg,
                                          NdviProcessor.PROCESSOR_NAME, JOptionPane.ERROR_MESSAGE);
        }
    }
}
