/*
 * $Id: NdviRequestElementFactory.java,v 1.2 2005/03/23 18:20:20 hartmut Exp $
 *
 * Copyright (C) 2004 by Brockmann Consult (info@brockmann-consult.de)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation. This program is distributed in the hope it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.bc.beam.processor.ndvi;

import org.esa.beam.framework.dataio.ProductIOPlugInManager;
import org.esa.beam.framework.param.ParamProperties;
import org.esa.beam.framework.param.ParamValidateException;
import org.esa.beam.framework.param.Parameter;
import org.esa.beam.framework.processor.*;
import org.esa.beam.util.StringUtils;

import java.io.File;
import java.net.URL;

/**
 * Responsible for creating the elements for a processing request used by the NDVI Processor.
 *
 * @author Sabine Embacher
 * @version $Revision: 1.2 $ $Date: 2005/03/23 18:20:20 $
 */
public class NdviRequestElementFactory implements RequestElementFactory {

    // singleton instance
    private static NdviRequestElementFactory _factory;

    private DefaultRequestElementFactory _defaultFactory;

    /**
     * Singleton interface - creates the one and only instance of this factory.
     * @return areference to the factory
     */
    public static NdviRequestElementFactory getInstance() {
        if (_factory == null) {
            _factory = new NdviRequestElementFactory();
        }
        return _factory;
    }

    /**
     * Creates an output format parameter.
     * @return
     */
    public Parameter createOutputFormatParameter() {
        ProductIOPlugInManager instance = ProductIOPlugInManager.getInstance();
        String[] formats = instance.getAllProductWriterFormatStrings();
        ParamProperties paramProperties = _defaultFactory.createStringParamProperties();
        if (formats.length > 0) {
            if (StringUtils.contains(formats, NdviProcessor.DEFAULT_OUTPUT_FORMAT)) {
                paramProperties.setDefaultValue(NdviProcessor.DEFAULT_OUTPUT_FORMAT);
            } else {
                paramProperties.setDefaultValue(formats[0]);
            }
        }
        paramProperties.setValueSet(formats);
        paramProperties.setValueSetBound(true);
        paramProperties.setReadOnly(true);
        paramProperties.setLabel(ProcessorConstants.OUTPUT_FORMAT_LABELTEXT);
        paramProperties.setDescription(ProcessorConstants.OUTPUT_FORMAT_DESCRIPTION);
        return new Parameter(ProcessorConstants.OUTPUT_FORMAT_PARAM_NAME, paramProperties);
    }


    public Parameter createParameter(String name, String value) throws RequestElementFactoryException {
        return _defaultFactory.createParameter(name, value);
    }

    /**
     * Creates a new reference to an input product for the current processing request.
     *
     * @param url the input product's URL, must not be <code>null</code>
     * @param fileFormat the file format, can be <code>null</code> if not known
     * @param typeId the product type identifier, can be <code>null</code> if not known
     * @throws IllegalArgumentException if <code>url</code> is <code>null</code>
     * @throws RequestElementFactoryException if the element could not be created
     */
    public ProductRef createInputProductRef(URL url, String fileFormat, String typeId) throws RequestElementFactoryException {
        return _defaultFactory.createInputProductRef(url, fileFormat, typeId);
    }

    /**
     * Creates a new reference to an output product for the current processing request.
     *
     * @param url the output product's URL, must not be <code>null</code>
     * @param fileFormat the file format, can be <code>null</code> if not known
     * @param typeId the product type identifier, can be <code>null</code> if not known
     * @throws IllegalArgumentException if <code>url</code> is <code>null</code>
     * @throws RequestElementFactoryException if the element could not be created
     */
    public ProductRef createOutputProductRef(URL url, String fileFormat, String typeId) throws RequestElementFactoryException {
        return _defaultFactory.createOutputProductRef(url, fileFormat, typeId);
    }

    /**
     * Creates a parameter for the default input product path - which is the
     * current user's home directory.
     */
    public Parameter createDefaultInputProductParameter() {
        return _defaultFactory.createDefaultInputProductParameter();
    }

    /**
     * Creates an output product parameter set to the default path.
     */
    public Parameter createDefaultOutputProductParameter() {
        Parameter defaultOutputProductParameter = _defaultFactory.createDefaultOutputProductParameter();
        ParamProperties properties = defaultOutputProductParameter.getProperties();
        Object defaultValue = properties.getDefaultValue();
        if (defaultValue instanceof File) {
            File file = (File) defaultValue;
            file = new File(file, NdviProcessor.DEFAULT_OUTPUT_DIR_NAME);
            properties.setDefaultValue(new File(file, NdviProcessor.DEFAULT_OUTPUT_PRODUCT_NAME));
        }
        defaultOutputProductParameter.setDefaultValue();
        return defaultOutputProductParameter;
    }

    /**
     * Creates a default logging pattern parameter set to the prefix passed in.
     * @param prefix the default setting for the logging pattern
     * @return a logging pattern Parameter conforming the system settings
     */
    public Parameter createDefaultLogPatternParameter(String prefix) {
        return _defaultFactory.createDefaultLogPatternParameter(prefix);
    }

    /**
     * Creates a logging to output product parameter set to true.
     * @return
     */
    public Parameter createLogToOutputParameter(String value) throws ParamValidateException {
        return _defaultFactory.createLogToOutputParameter(value);
    }

    ///////////////////////////////////////////////////////////////////////////
    /////// END OF PUBLIC
    ///////////////////////////////////////////////////////////////////////////

    /**
     * Constructs the object.
     */
    private NdviRequestElementFactory() {
        _defaultFactory = DefaultRequestElementFactory.getInstance();
    }
}
